package parser

import "git.tebibyte.media/sashakoshka/fspl/lexer"
import "git.tebibyte.media/sashakoshka/fspl/errors"
import "git.tebibyte.media/sashakoshka/fspl/entity"

var descriptionExpression = "Expression"
var startTokensExpression = []lexer.TokenKind {
	lexer.Ident,
	lexer.LBracket,
	lexer.LParen,
	lexer.Int,
	lexer.Float,
	lexer.String,
}

func (this *Parser) parseExpression () (entity.Expression, error) {
	// TODO
	// 1. Run a decision tree to parse the expression.
	// 2. After that, test for infix operators (. and =). We will not need
	//    to consume more tokens to do this, as if they exist they will be
	//    the very next token AKA the current one after parsing the
	//    expression.
	// 3. Then, use the previously parsed expression and use it as the left
	//    side, and go on to parse the right side. Assignment (=) is the
	//    only infix expression where both left and right sides are
	//    expressions, so it gets to call parseExpression() and recurse. The
	//    implied associativity matters here.
	// 4. If not assignment, then this has to do with member access (the
	//    right side is not an expression), so we repeat from step 2 in order
	//    to chain member accesses and method calls. Step 2 will test
	//    whether or not we need to continue repeating afterward. The
	//    implied associativity matters here.
	// 5. Return the expression, whether it was directly parsed or is a tree
	//    of infix operator(s).
	expression, err := this.parseExpressionRoot()
	if err != nil { return nil, err }
	
	return expression, nil
}

func (this *Parser) parseExpressionRoot () (entity.Expression, error) {
	err := this.expectDesc(descriptionExpression, startTokensExpression...)
	if err != nil { return nil, err }
	
	switch this.kind() {
	case lexer.Ident:    return this.parseExpressionRootIdent()
	case lexer.LBracket: // TODO
	case lexer.LParen:   // TODO
	case lexer.Int:      return this.parseLiteralInt()
	case lexer.Float:    return this.parseLiteralFloat()
	case lexer.String:   return this.parseLiteralString()
	}
	panic(this.bug())
}

func (this *Parser) parseExpressionRootIdent () (entity.Expression, error) {
	err := this.expect(lexer.Ident)
	if err != nil { return nil, err }
	name := this.value()
	
	switch name {
	case "true", "false": return this.parseLiteralBoolean ()
	case "nil":           return this.parseLiteralNil()
	case "if":            // TODO
	case "loop":          // TODO
	default:              return this.parseVariableOrDeclaration()
	}
	panic(this.bug())
}

var descriptionDeclaration = "Declaration"
var startTokensDeclaration = []lexer.TokenKind { lexer.Ident }

func (this *Parser) parseDeclaration () (*entity.Declaration, error) {
	err := this.expectDesc(descriptionDeclaration, startTokensDeclaration...)
	if err != nil { return nil, err }
	name := this.value()
	pos  := this.pos()
	this.next()
	return this.parseDeclarationCore(pos, name)
}

func (this *Parser) parseDeclarationCore (pos errors.Position, name string) (*entity.Declaration, error) {
	err := this.expect(lexer.Colon)
	if err != nil { return nil, err }
	this.next()
	ty, err := this.parseType()
	if err != nil { return nil, err }
	
	return &entity.Declaration {
		Position: pos,
		Name:     name,
		Ty:       ty,
	}, nil
}

func (this *Parser) parseVariableOrDeclaration () (entity.Expression, error) {
	err := this.expectDesc("Variable or Declaration", lexer.Ident)
	if err != nil { return nil, err }

	variable := &entity.Variable {
		Position: this.pos(),
		Name:     this.value(),
	}
	
	this.next()
	if this.token.Is(lexer.Colon) {
		return this.parseDeclarationCore(variable.Position, variable.Name)
	} else {
		return variable, nil
	}
}

func (this *Parser) parseLiteralInt () (*entity.LiteralInt, error) {
	err := this.expect(lexer.Int)
	if err != nil { return nil, err }
	defer this.next()
	
	value, err := parseInteger(this.value())
	if err != nil { return nil, errors.Errorf(this.pos(), err.Error()) }
	
	return &entity.LiteralInt {
		Position: this.pos(),
		Value:    int(value), // TODO struct should store as int64
	}, nil
}

func (this *Parser) parseLiteralFloat () (*entity.LiteralFloat, error) {
	err := this.expect(lexer.Float)
	if err != nil { return nil, err }
	defer this.next()
	
	value, err := parseFloat(this.value())
	if err != nil { return nil, errors.Errorf(this.pos(), err.Error()) }
	
	return &entity.LiteralFloat {
		Position: this.pos(),
		Value:    value,
	}, nil
}

func (this *Parser) parseLiteralString () (*entity.LiteralString, error) {
	err := this.expect(lexer.String)
	if err != nil { return nil, err }
	defer this.next()
	
	return &entity.LiteralString {
		Position:  this.pos(),
		ValueUTF8: this.value(),
	}, nil
}

func (this *Parser) parseLiteralBoolean () (*entity.LiteralBoolean, error) {
	err := this.expectValueDesc("Boolean", lexer.Ident, "true", "false")
	if err != nil { return nil, err }
	defer this.next()
	
	return &entity.LiteralBoolean {
		Position: this.pos(),
		Value:    this.value() == "true",
	}, nil
}

func (this *Parser) parseLiteralNil () (*entity.LiteralNil, error) {
	err := this.expectValueDesc("Nil", lexer.Ident, "nil")
	if err != nil { return nil, err }
	defer this.next()
	
	return &entity.LiteralNil {
		Position: this.pos(),
	}, nil
}
